@@ -21,7 +21,9 @@
  */
 
 #include <string.h>
+#include <stdlib.h>
 #include <popt.h>
+#include <sys/time.h>
 #include <gdk/gdkwindow.h>
 #include <gdk/gdkx.h>
 #include <gtk/gtk.h>
@@ -33,6 +35,8 @@
 
 #define ENV_STRING_MAX_SIZE 128
 
+GNOME_Magnifier_ZoomRegion zoom_region;
+
 typedef struct {
 	gchar *target_display;
 	gchar *source_display;
@@ -52,7 +56,11 @@ typedef struct {
 	int   mouse_follow;
 	int   invert_image;
 	int   no_initial_region;
-	int   timing_test;
+	int   timing_iterations;
+	int   timing_output;
+	int   timing_delta_x;
+	int   timing_delta_y;
+	int   timing_pan_rate;
 	int   smooth_scroll;
 	int   border_width;
 	unsigned long  border_color;
@@ -79,6 +87,10 @@ static MagnifierOptions global_options = { ":0.0",
 					   0,
 					   0,
 					   0,
+					   10,
+					   10,
+					   0,
+					   0,
 					   0,
 					   0,
 					   0
@@ -99,9 +111,13 @@ struct poptOption magnifier_options [] = {
 	{"zoom-factor", 'z', POPT_ARG_FLOAT, &global_options.zoom_factor, 'z', "zoom (scale) factor used to magnify source display", NULL}, 
 	{"invert-image", 'i', POPT_ARG_NONE, &global_options.invert_image, 'i', "invert the image colormap", NULL}, 
 	{"no-initial-region", '\0', POPT_ARG_NONE, &global_options.no_initial_region, '\0', "don't create an initial zoom region", NULL},
-	{"timing-test", '\0', POPT_ARG_NONE, &global_options.timing_test, '\0', "carry out timing benchmark test", NULL},
+	{"timing-iterations", '\0', POPT_ARG_INT, &global_options.timing_iterations, '\0', "iterations to run timing benchmark test (0=continuous)", NULL},
+	{"timing-output", '\0', POPT_ARG_NONE, &global_options.timing_output, '\0', "display performance ouput", NULL},
+	{"timing-pan-rate", '\0', POPT_ARG_INT, &global_options.timing_pan_rate, '\0', "timing pan rate in lines per frame", NULL},
+	{"timing-delta-x", '\0', POPT_ARG_INT, &global_options.timing_delta_x, '\0', "pixels to pan in x-dimension each frame in timing update test", NULL},
+	{"timing-delta-y", '\0', POPT_ARG_INT, &global_options.timing_delta_y, '\0', "pixels to pan in y-dimension each frame in timing update test", NULL},
 	{"smoothing-type", '\0', POPT_ARG_STRING, &global_options.smoothing_type, '\0', "image smoothing algorithm to apply (bilinear-interpolation | none)", NULL},
-	{"fullscreen", 'f', POPT_ARG_NONE, &global_options.fullscreen, '\0', "fullscreen magnification, covers entire target display", NULL},
+	{"fullscreen", 'f', POPT_ARG_NONE, &global_options.fullscreen, '\0', "fullscreen magnification, covers entire target display [REQUIRES --source-display and --target-display]", NULL},
 	{"smooth-scrolling", '\0', POPT_ARG_NONE, &global_options.smooth_scroll, '\0', "use smooth scrolling", NULL},
 	{"border-size", 'b', POPT_ARG_INT, &global_options.border_width, '\0', "width of border", NULL},
 	{"border-color", 'c', POPT_ARG_LONG, &global_options.border_color, '\0', "border color specified as (A)RGB 23-bit value, Alpha-MSB", NULL},
@@ -120,39 +136,92 @@ init_rect_bounds (GNOME_Magnifier_RectBounds *bounds,
 	bounds->y2 = y2;
 }
 
+static int screen_width, screen_height;
+
 static int
 magnifier_main_test_image (gpointer data)
 {
-	static int test_i_foo = 0;
+	static long timing_counter = 0;
+	static int timing_x_pos = 0;
+	static int timing_y_pos = 0;
+	static int x_direction = 1;
+	static int y_direction = 1;
 	Magnifier *magnifier = (Magnifier *) data;
 	GNOME_Magnifier_ZoomRegionList *zoom_regions;
+	Bonobo_PropertyBag properties;
 	CORBA_Environment ev;
 	GNOME_Magnifier_RectBounds roi;
+	int x_roi, y_roi;
+
+	/* Only iterate the number of times specified */
+	if (global_options.timing_iterations > 0) {
+		if (timing_counter > global_options.timing_iterations) {
+  			CORBA_exception_init (&ev);
+	  		properties = GNOME_Magnifier_ZoomRegion_getProperties (zoom_region, &ev);
+			if (BONOBO_EX (&ev))
+				fprintf (stderr, "EXCEPTION\n");
+
+	  		bonobo_pbclient_set_boolean (properties, "exit-magnifier",
+				       TRUE, &ev);
+		}
+	}
+
 	CORBA_exception_init (&ev);
 
-	roi.x1 = 10 * test_i_foo;
-	roi.y1 = 10 * test_i_foo;
-	roi.x2 = 200 + 10 * test_i_foo;
-	roi.y2 = 200 + 10 * test_i_foo;
-	++test_i_foo;
+	x_roi  = global_options.timing_delta_x * timing_x_pos;
+	roi.x1 = x_roi;
+	roi.x2 = (screen_width / global_options.zoom_factor) + roi.x1;
+	x_roi  = global_options.timing_delta_x * (timing_x_pos + x_direction);
+
+	/* Determine if magnifier hit an edge and should reverse direction */
+	if (x_roi + (screen_width / global_options.zoom_factor) > screen_width)
+		x_direction = -1;
+	else if (x_roi < 0)
+		x_direction = 1;
+
+	timing_x_pos += x_direction;
+
+	y_roi = global_options.timing_delta_y * timing_y_pos;
+
+	/* Calculate size of screen not covered by magnifier */
+	if (global_options.horizontal_split)
+		roi.y1 = y_roi + screen_height;
+	else
+		roi.y1 = y_roi;
+	roi.y2 = (screen_height / global_options.zoom_factor) + roi.y1;
+
+	y_roi = global_options.timing_delta_y * (timing_y_pos + y_direction);
+
+	/* The counter is increased each time the y-direction changes */
+	if (y_roi + (screen_height / global_options.zoom_factor) > screen_height) {
+		timing_counter++;
+		y_direction = -1;
+	}
+	else if (y_roi < 0) {
+		timing_counter++;
+		y_direction = 1;
+	}
+
+	timing_y_pos += y_direction;
 
 	if (!IS_MAGNIFIER (magnifier))
 		return FALSE;
+
 	magnifier->priv->cursor_x = (roi.x2 + roi.x1) / 2;
 	magnifier->priv->cursor_y = (roi.y2 + roi.y1) / 2;
+
 	zoom_regions =
 		GNOME_Magnifier_Magnifier_getZoomRegions (
 			BONOBO_OBJREF (magnifier),
 			&ev);
 
-	if (zoom_regions && (zoom_regions->_length > 0))
+	if (zoom_regions && (zoom_regions->_length > 0)) {
+
 		GNOME_Magnifier_ZoomRegion_setROI (
 			zoom_regions->_buffer[0], &roi, &ev);
-  
-	if (test_i_foo < 50)
-		return TRUE;
+	}
 
-	return FALSE;
+	return TRUE;
 }
 
 static int last_x = 0, last_y = 0;
@@ -174,7 +243,7 @@ magnifier_main_pan_image (gpointer data)
   if (global_options.mouse_follow && IS_MAGNIFIER (magnifier))
   {
 	  gdk_window_get_pointer (
-		  magnifier->priv->root,
+		  magnifier_get_root (magnifier),
 		  &mouse_x_return,
 		  &mouse_y_return,
 		  &mask_return);
@@ -271,25 +340,18 @@ main (int argc, char** argv)
   GNOME_Magnifier_RectBounds *roi = GNOME_Magnifier_RectBounds__alloc();
   GNOME_Magnifier_RectBounds *viewport = GNOME_Magnifier_RectBounds__alloc();
   CORBA_any *viewport_any;
-  int w, h, x = 0, y = 0;
+  int x = 0, y = 0;
   guint pan_handle = 0, refresh_handle = 0;
   CORBA_Environment ev;
   Bonobo_PropertyBag properties;
 
   Magnifier *magnifier;
   
-  if (!bonobo_init (&argc, argv))
-    {
-      g_error ("Could not initialize Bonobo");
-    }
+  if (!bonobo_init (&argc, argv)) {
+    g_error ("Could not initialize Bonobo");
+  }
   CORBA_exception_init (&ev);
   
-  /* FIXME */
-  gtk_init (&argc, &argv);
-
-  magnifier = magnifier_new (argc, argv);
-  fprintf (stderr, "New magnifier created.\n");
-  
   ctx = poptGetContext ("magnifier",
 			argc,
 			(const char **)argv,
@@ -299,6 +361,28 @@ main (int argc, char** argv)
   while (poptGetNextOpt (ctx) >= 0)
 	/**/;
 
+  /** 
+   * slight hack, we set $DISPLAY if --target-display is set.
+   * this is because I don't think GTK+ works well with across multiple 
+   * servers yet.
+   **/
+  if (global_options.target_display) {
+	  gchar *string;
+	  string = g_strconcat ("DISPLAY=", global_options.target_display, NULL);
+	  putenv (string);
+  }
+
+  if (global_options.timing_pan_rate && global_options.timing_iterations == 0)
+  {
+    g_error ("Must specify timing_iterations when running pan test");
+  }
+
+  /* FIXME */
+  gtk_init (&argc, &argv);
+
+  magnifier = magnifier_new ();
+  fprintf (stderr, "New magnifier created.\n");
+  
   properties = GNOME_Magnifier_Magnifier_getProperties (
 	  BONOBO_OBJREF (magnifier), &ev);
   if (ev._major != CORBA_NO_EXCEPTION) fprintf (stderr, "EXCEPTION\n");
@@ -327,26 +411,27 @@ main (int argc, char** argv)
 	  bonobo_pbclient_set_ulong (properties, "cursor-color",
 				     global_options.cursor_color, 
 				     NULL);
-  w = gdk_screen_get_width (
+
+  screen_width = gdk_screen_get_width (
 	  gdk_display_get_screen (magnifier->target_display,
 				  magnifier->target_screen_num));
-  h = gdk_screen_get_height (
+  screen_height = gdk_screen_get_height (
 	  gdk_display_get_screen (magnifier->target_display,
 				  magnifier->target_screen_num));
 
   if (global_options.vertical_split) {
-	  w /= 2;
-	  x = w;
+	  screen_width /= 2;
+	  x = screen_width;
   }
   if (global_options.horizontal_split) {
-	  h /= 2;
+	  screen_height /= 2;
   }
 
-  fprintf (stderr, "initial viewport %d %d\n", (int) w,
-	   (int) h);
+  fprintf (stderr, "initial viewport %d %d\n", (int) screen_width,
+	   (int) screen_height);
 
   viewport_any = CORBA_any__alloc ();
-  init_rect_bounds (viewport, x, y, x + w, y + h);
+  init_rect_bounds (viewport, x, y, x + screen_width, y + screen_height);
   viewport_any->_type = TC_GNOME_Magnifier_RectBounds;
   viewport_any->_value = ORBit_copy_value (viewport,
 					   TC_GNOME_Magnifier_RectBounds);
@@ -363,11 +448,10 @@ main (int argc, char** argv)
       global_options.fullscreen)
   {
 	  int scroll_policy;
-	  GNOME_Magnifier_ZoomRegion zoom_region;
 	  
 	  fprintf (stderr, "creating an initial zoom region.\n");
 	  init_rect_bounds (roi, 0, 0, 100, 100);
-	  init_rect_bounds (viewport, 0, 0, w, h);
+	  init_rect_bounds (viewport, 0, 0, screen_width, screen_height);
 	  zoom_region =
 		  GNOME_Magnifier_Magnifier_createZoomRegion (
 			  BONOBO_OBJREF (magnifier),
@@ -377,14 +461,20 @@ main (int argc, char** argv)
 			  viewport,
 			  &ev);
 	  
-	  /* XXX FIXME: TODO: free the property-sets */
 	  properties = GNOME_Magnifier_ZoomRegion_getProperties (zoom_region, &ev);
-	  scroll_policy = global_options.smooth_scroll ?
-		  GNOME_Magnifier_ZoomRegion_SCROLL_SMOOTHEST :
-		  GNOME_Magnifier_ZoomRegion_SCROLL_FASTEST;
 	  if (BONOBO_EX (&ev))
 		  fprintf (stderr, "EXCEPTION\n");
 
+	  scroll_policy = global_options.smooth_scroll ?
+		  GNOME_Magnifier_ZoomRegion_SCROLL_SMOOTHEST :
+		  GNOME_Magnifier_ZoomRegion_SCROLL_FASTEST;
+
+	  bonobo_pbclient_set_long (properties, "timing-iterations",
+				       global_options.timing_iterations, &ev);
+	  bonobo_pbclient_set_boolean (properties, "timing-output",
+				       global_options.timing_output, &ev);
+	  bonobo_pbclient_set_long (properties, "timing-pan-rate",
+				       global_options.timing_pan_rate, &ev);
 	  bonobo_pbclient_set_boolean (properties, "smooth-scroll-policy",
 				       scroll_policy, &ev);
 	  bonobo_pbclient_set_long    (properties, "border_size",
@@ -411,25 +501,48 @@ main (int argc, char** argv)
 	  properties = NULL;
   }
 
-  if (!global_options.timing_test) {
-	  refresh_handle = gtk_timeout_add (
+  if (global_options.timing_pan_rate)
+  {
+	  GNOME_Magnifier_ZoomRegionList *zoom_regions;
+	  GNOME_Magnifier_RectBounds roi;
+	  roi.x1 = 100;
+	  roi.x2 = 100 + (screen_width / global_options.zoom_factor);
+	  roi.y1 = 0;
+	  roi.y2 = screen_height / global_options.zoom_factor;
+	  
+	  zoom_regions = GNOME_Magnifier_Magnifier_getZoomRegions (
+		  BONOBO_OBJREF (magnifier), &ev);
+	  
+	  if (zoom_regions && (zoom_regions->_length > 0))
+	  {
+		  GNOME_Magnifier_ZoomRegion_setROI (
+			  zoom_regions->_buffer[0], &roi, &ev);
+	  }
+  }
+  else if (global_options.timing_iterations)
+  {
+	  refresh_handle = g_timeout_add (global_options.refresh_time,
+					  magnifier_main_test_image,
+					  magnifier);
+  }
+  else
+  {
+	  refresh_handle = g_timeout_add (
 		  global_options.refresh_time,
 		  magnifier_main_refresh_all, magnifier);
-
-	  pan_handle = gtk_timeout_add (
+	  
+	  pan_handle = g_timeout_add (
 		  global_options.mouse_poll_time,
 		  magnifier_main_pan_image, magnifier);
-  } else
-	  refresh_handle = gtk_timeout_add (
-		  global_options.refresh_time,
-		  magnifier_main_test_image, magnifier);
-
+  }
+  
   bonobo_main ();
   
-  gtk_timeout_remove (refresh_handle);
-
+  if (refresh_handle)
+	  g_source_remove (refresh_handle);
+  
   if (pan_handle)
-	  gtk_timeout_remove (pan_handle);
+	  g_source_remove (pan_handle);
   
   return 0;
 }
